/*
 * Copyright (c) 2003 Shaven Puppy Ltd
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of 'Shaven Puppy' nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package riven.classpath;

import java.io.UnsupportedEncodingException;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

public class Text {

	public static boolean contains(String value, char find) {
		return value.indexOf(find) != -1;
	}

	public static int count(String value, char find) {
		int matches = 0;
		for (int i = 0; i < value.length(); i++) {
			if (value.charAt(i) == find) {
				matches++;
			}
		}
		return matches;
	}

	public static String remove(String value, char find) {
		char[] hold = new char[value.length() - count(value, find)];
		for (int i = 0, k = 0; i < value.length(); i++) {
			if (value.charAt(i) != find) {
				hold[k++] = value.charAt(i);
			}
		}
		return new String(hold);
	}

	public static String removeWhitespace(String value) {
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < value.length(); i++) {
			if (!Character.isWhitespace(value.charAt(i))) {
				sb.append(value.charAt(i));
			}
		}
		return sb.toString();
	}

	// ascii

	public static final String ascii(byte[] raw, int off, int len) {
		char[] str = new char[len];
		for (int i = 0; i < len; i++) {
			str[i] = (char) raw[off + i];
		}
		return new String(str);
	}

	public static final String ascii(byte[] raw) {
		char[] str = new char[raw.length];
		for (int i = 0; i < str.length; i++) {
			str[i] = (char) raw[i];
		}
		return new String(str);
	}

	public static final byte[] ascii(String str) {
		byte[] raw = new byte[str.length()];
		for (int i = 0; i < raw.length; i++) {
			raw[i] = (byte) str.charAt(i);
		}
		return raw;
	}

	// utf8

	public static final String utf8(byte[] raw) {
		return Text.utf8(raw, 0, raw.length);
	}

	public static final String utf8(byte[] raw, int off, int len) {
		try {
			return new String(raw, off, len, "UTF-8");
		} catch (UnsupportedEncodingException exc) {
			throw new IllegalStateException();
		}
	}

	public static final byte[] utf8(String str) {
		try {
			return str.getBytes("UTF-8");
		} catch (UnsupportedEncodingException exc) {
			throw new IllegalStateException();
		}
	}

	// between

	public static final String between(String value, char start, char end) {
		value = after(value, start);
		return (value == null) ? null : before(value, end);
	}

	public static final String between(String value, String start, String end) {
		value = after(value, start);
		return (value == null) ? null : before(value, end);
	}

	// before (String, char)

	public static final String before(String value, char find) {
		return before(value, find, null);
	}

	public static final String beforeLast(String value, char find) {
		return beforeLast(value, find, null);
	}

	public static final String before(String value, char find, String defaultValue) {
		int indexOf = value.indexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(0, indexOf);
	}

	public static final String beforeLast(String value, char find, String defaultValue) {
		int indexOf = value.lastIndexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(0, indexOf);
	}

	// before (String, String)

	public static final String before(String value, String find) {
		return before(value, find, null);
	}

	public static final String beforeLast(String value, String find) {
		return beforeLast(value, find, null);
	}

	public static final String before(String value, String find, String defaultValue) {
		int indexOf = value.indexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(0, indexOf);
	}

	public static final String beforeLast(String value, String find, String defaultValue) {
		int indexOf = value.lastIndexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(0, indexOf);
	}

	// after (String, char)

	public static final String after(String value, char find) {
		return after(value, find, null);
	}

	public static final String afterLast(String value, char find) {
		return afterLast(value, find, null);
	}

	public static final String after(String value, char find, String defaultValue) {
		int indexOf = value.indexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(indexOf + 1);
	}

	public static final String afterLast(String value, char find, String defaultValue) {
		int indexOf = value.lastIndexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(indexOf + 1);
	}

	// after (String, String)

	public static final String after(String value, String find) {
		return after(value, find, null);
	}

	public static final String afterLast(String value, String find) {
		return afterLast(value, find, null);
	}

	public static final String after(String value, String find, String defaultValue) {
		int indexOf = value.indexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(indexOf + find.length());
	}

	public static final String afterLast(String value, String find, String defaultValue) {
		int indexOf = value.lastIndexOf(find);
		return (indexOf == -1) ? defaultValue : value.substring(indexOf + find.length());
	}

	// split

	public static final String[] splitPair(String input, char delim) {
		int io = input.indexOf(delim);
		if (io == -1) {
			return null;
		}
		return new String[] { input.substring(0, io), input.substring(io + 1) };
	}

	public static final List<String> split(String input, char delim) {
		List<String> parts = new ArrayList<>();
		for (int index; (index = input.indexOf(delim)) != -1;) {
			parts.add(input.substring(0, index));
			input = input.substring(index + 1);
		}
		parts.add(input);
		return parts;
	}

	public static final String join(List<String> values, char delim) {
		int len = Math.max(0, values.size() - 1);
		for (int i = 0; i < values.size(); i++) {
			len += values.get(i).length();
		}

		StringBuilder sb = new StringBuilder(len);
		for (String value : values) {
			if (sb.length() != 0) {
				sb.append(delim);
			}
			sb.append(value);
		}
		return sb.toString();
	}

	public static final String join(String[] values, char delim) {
		int len = Math.max(0, values.length - 1);
		for (int i = 0; i < values.length; i++) {
			len += values[i].length();
		}

		StringBuilder sb = new StringBuilder(len);
		for (String value : values) {
			if (sb.length() != 0) {
				sb.append(delim);
			}
			sb.append(value);
		}
		return sb.toString();
	}

	public static final String join(String[] values, String delim) {
		int len = Math.max(0, (values.length - 1) * delim.length());
		for (int i = 0; i < values.length; i++) {
			len += values[i].length();
		}

		StringBuilder sb = new StringBuilder(len);
		for (String value : values) {
			if (sb.length() != 0) {
				sb.append(delim);
			}
			sb.append(value);
		}
		return sb.toString();
	}

	public static final String join(List<String> values) {
		int len = 0;
		for (int i = 0; i < values.size(); i++) {
			len += values.get(i).length();
		}

		StringBuilder sb = new StringBuilder(len);
		for (String value : values) {
			sb.append(value);
		}
		return sb.toString();
	}

	public static final String join(String... values) {
		int len = 0;
		for (int i = 0; i < values.length; i++) {
			len += values[i].length();
		}

		StringBuilder sb = new StringBuilder(len);
		for (String value : values) {
			sb.append(value);
		}
		return sb.toString();
	}

	// gen

	private static final char[] generatecode_default_chars;
	static {
		generatecode_default_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz01234567890".toCharArray();
	}

	public static String generateRandomCode(int len) {
		return generateRandomCode(len, generatecode_default_chars, new SecureRandom());
	}

	public static String generateRandomCode(int len, char[] chars, Random r) {
		char[] dst = new char[len];
		for (int i = 0; i < dst.length; i++) {
			dst[i] = chars[r.nextInt(chars.length)];
		}
		return new String(dst);
	}
}