package riven.classpath.ui;

import java.awt.Component;
import java.awt.Container;
import java.awt.Window;
import java.util.NoSuchElementException;

public class DOM {
	public static <T extends Component> T getByID(Component root, String id) {
		T got = searchByID(root, id);
		if (got == null) {
			throw new NoSuchElementException(id);
		}
		return got;
	}

	@SuppressWarnings("unchecked")
	public static <T extends Component> T searchByID(Component root, String id) {

		if (id.equals(root.getName())) {
			return (T) root;
		}

		if (root instanceof Window) {
			for (Window window : ((Window) root).getOwnedWindows()) {
				T got = searchByID(window, id);
				if (got != null) {
					return got;
				}
			}
		}

		if (root instanceof Container) {
			for (Component child : ((Container) root).getComponents()) {
				T got = searchByID(child, id);
				if (got != null) {
					return got;
				}
			}
		}

		return null;
	}

	public static void setID(Component c, String id) {
		if (searchByID(getRoot(c), id) != null) {
			throw new IllegalStateException("id [" + id + "] already exists");
		}

		c.setName(id);
	}

	private static Container getRoot(Component c) {
		Component p = c;
		while (p.getParent() != null) {
			p = p.getParent();
		}
		return (Container) p;
	}
}
