package riven.classpath.ui;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import riven.classpath.Producer;
import riven.classpath.StateCallback;
import riven.classpath.Text;

public class UserIO {
	private static boolean is_lnf_set = false;

	public static synchronized void setSystemLookAndFeel() {
		if (is_lnf_set) {
			return;
		}
		is_lnf_set = true;

		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception exc) {
			exc.printStackTrace();
		}
	}

	public static void showFrame(JFrame frame, int actionOnExit) {
		frame.setResizable(true);
		frame.pack();
		frame.setLocationRelativeTo(null);
		frame.setDefaultCloseOperation(actionOnExit);
		frame.setVisible(true);
	}

	public static void showDialog(JDialog dialog, int actionOnExit) {
		dialog.setResizable(true);
		dialog.pack();
		dialog.setLocationRelativeTo(null);
		dialog.setDefaultCloseOperation(actionOnExit);
		dialog.setVisible(true);
	}

	public static void alert(String title, String msg) {
		JOptionPane.showMessageDialog(null, msg, title, JOptionPane.WARNING_MESSAGE);
	}

	public static <T> void async(final Producer<T> producer, final StateCallback<T> callback) {
		new Thread(new Runnable() {
			@Override
			public void run() {
				try {
					final T got = producer.produce();

					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							callback.onSuccess(got);
						}
					});

				} catch (final Exception exc) {
					SwingUtilities.invokeLater(new Runnable() {
						@Override
						public void run() {
							callback.onFailure(exc);
						}
					});
				}
			}
		}).start();
	}

	public static void setClipboard(final String contents) {
		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

		clipboard.setContents(new Transferable() {

			@Override
			public boolean isDataFlavorSupported(DataFlavor flavor) {
				return flavor.isFlavorTextType();
			}

			@Override
			public DataFlavor[] getTransferDataFlavors() {
				try {
					return new DataFlavor[] { new DataFlavor("text/plain") };
				} catch (Exception exc) {
					throw new IllegalStateException(exc);
				}
			}

			@Override
			public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
				return new ByteArrayInputStream(Text.ascii(contents));
			}
		}, new ClipboardOwner() {
			@Override
			public void lostOwnership(Clipboard clipboard, Transferable contents) {
			}
		});
	}
}
