package riven.classpath.ui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.List;
import java.util.jar.*;

import javax.swing.*;
import javax.swing.event.*;

import riven.classpath.*;

public class WhereIsThatStaticImport {
	public static void main(String[] args) throws IOException {

		final String entryPrefix = "org/lwjgl/opengl/";
		WhereIsThatStaticImport whtsi = null;

		String jar;
		if (args.length == 1) {
			// user provided the location...
			jar = args[0];
			File jarFile = new File(jar);
			if (!jarFile.exists()) {
				System.out.println("file not found: " + jar);
				return;
			}

			System.out.println("loading file: " + jarFile.getAbsolutePath());
			addClassPathEntry(jarFile.toURI().toURL());

			whtsi = new WhereIsThatStaticImport();
			try (InputStream in = new FileInputStream(jarFile)) {
				whtsi.discoverJar(in, entryPrefix);
			} catch (IOException exc) {
				exc.printStackTrace();
			}

			jar = jar.replace('\\', '/');
			if (jar.contains("/")) {
				jar = Text.afterLast(jar, '/');
			}
		} else {
			jar = "lwjgl.jar";

			// find it in the classpath then...
			for (URL url : getClassPath()) {
				System.out.println("classpath entry: " + url);
				if (!url.getFile().endsWith("/" + jar)) {
					continue;
				}

				whtsi = new WhereIsThatStaticImport();
				try (InputStream in = url.openStream()) {
					whtsi.discoverJar(in, entryPrefix);
				} catch (IOException exc) {
					exc.printStackTrace();
				}
				break;
			}

			if (whtsi == null) {
				System.out.println("failed to find " + jar + " in classpath.");
				System.exit(0);
			}
		}

		final WhereIsThatStaticImport dataSource = whtsi;

		//

		UserIO.setSystemLookAndFeel();

		final JFrame frame = new JFrame();
		frame.setTitle("Static import - " + jar);

		JTextField queryField = new JTextField();
		DOM.setID(queryField, "query-field");

		JList<String> resultList = new JList<String>();
		resultList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		DOM.setID(resultList, "result-list");

		JTextField importField = new JTextField();
		importField.setFont(new Font(Font.MONOSPACED, Font.PLAIN, importField.getFont().getSize()));
		importField.setEditable(false);
		importField.setBackground(Color.WHITE);
		importField.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				this.selectAll();
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				this.selectAll();
			}

			private void selectAll() {
				JTextField importField = DOM.getByID(frame, "import-field");

				// select everything on click
				importField.setSelectionStart(0);
				importField.setSelectionEnd(importField.getText().length());
			}
		});
		DOM.setID(importField, "import-field");

		//

		final Ref<List<Pair<Class<?>, ?>>> queryResult = new Ref<>();

		resultList.addListSelectionListener(new ListSelectionListener() {
			@Override
			public void valueChanged(ListSelectionEvent e) {
				JList<String> resultList = DOM.getByID(frame, "result-list");
				JTextField importField = DOM.getByID(frame, "import-field");

				int[] indices = resultList.getSelectedIndices();
				int index = indices.length == 0 ? -1 : indices[0];

				if (index == -1 || index >= queryResult.get().size()) {
					importField.setText("");
				} else {
					Pair<Class<?>, ?> item = queryResult.get().get(index);
					String className = item.first().getName();
					String identifierName = getIdentifierName(item);
					importField.setText("import static " + className + "." + identifierName + ";");
				}
			}
		});

		queryField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				JList<String> resultList = DOM.getByID(frame, "result-list");

				int[] indices = resultList.getSelectedIndices();
				int oldIndex = indices.length == 0 ? -1 : indices[0];
				int newIndex = oldIndex;

				if (e.getKeyCode() == KeyEvent.VK_UP) {
					newIndex -= 1;
				} else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
					newIndex += 1;
				}

				if (oldIndex != newIndex) {
					int size = queryResult.get() == null ? 0 : queryResult.get().size();
					newIndex = Math.min(size - 1, Math.max(0, newIndex));
					if (newIndex != -1) {
						resultList.setSelectedIndex(newIndex);
					}
				}
			}
		});

		queryField.getDocument().addDocumentListener(new DocumentAdapter() {
			@Override
			public void insertUpdate(DocumentEvent e) {
				this.performQuery();
			}

			@Override
			public void removeUpdate(DocumentEvent e) {
				this.performQuery();
			}

			@Override
			public void changedUpdate(DocumentEvent e) {
				this.performQuery();
			}

			private void performQuery() {
				JTextField queryField = DOM.getByID(frame, "query-field");
				JList<String> resultList = DOM.getByID(frame, "result-list");

				String query = queryField.getText().trim();
				System.out.println("query: \"" + query + "\"");

				List<Pair<Class<?>, ?>> holder = new ArrayList<>();

				int maxResults = 0;

				if (query.isEmpty()) {
					// do nothing
				} else if (Character.isUpperCase(query.charAt(0))) {
					query = query.toUpperCase();
					for (Pair<Class<?>, ?> pair : dataSource.queryField(query, maxResults)) {
						holder.add(pair);
					}
				} else {
					for (Pair<Class<?>, ?> pair : dataSource.queryMethod(query, maxResults)) {
						holder.add(pair);
					}
				}
				queryResult.set(holder);

				AbstractListModel<String> listModel = new AbstractListModel<String>() {
					@Override
					public String getElementAt(int index) {
						Pair<Class<?>, ?> item = queryResult.get().get(index);
						return item.first().getSimpleName() + "." + getIdentifierName(item);
					}

					@Override
					public int getSize() {
						return queryResult.get().size();
					}
				};

				resultList.setModel(listModel);

				if (!queryResult.get().isEmpty()) {
					resultList.setSelectedIndex(0);
				}
			}
		});
		//

		JPanel queryPanel = new JPanel(new BorderLayout(0, 2)) {
			@Override
			public Dimension getPreferredSize() {
				Dimension dim = super.getPreferredSize();
				dim.width = Math.max(dim.width, 512);
				return dim;
			}
		};
		queryPanel.setBorder(BorderUtil.empty(8));
		queryPanel.add(new JLabel("Method name or field name:"), BorderLayout.NORTH);
		queryPanel.add(queryField, BorderLayout.SOUTH);

		JPanel resultPanel = new JPanel(new BorderLayout(0, 2)) {
			@Override
			public Dimension getPreferredSize() {
				Dimension dim = super.getPreferredSize();
				dim.width = Math.max(dim.width, 512);
				return dim;
			}
		};
		resultPanel.setBorder(BorderUtil.empty(8));
		resultPanel.add(new JLabel("Results:"), BorderLayout.NORTH);
		resultPanel.add(new JScrollPane(resultList), BorderLayout.CENTER);
		resultPanel.add(importField, BorderLayout.SOUTH);

		//

		{
			JPanel cp = (JPanel) frame.getContentPane();
			cp.setLayout(new BorderLayout());
			cp.add(queryPanel, BorderLayout.NORTH);
			cp.add(resultPanel, BorderLayout.CENTER);
		}

		frame.setAlwaysOnTop(true);
		frame.setResizable(false);
		frame.pack();
		frame.setLocationRelativeTo(null);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);
	}

	private static URL[] getClassPath() {
		return ((URLClassLoader) WhereIsThatStaticImport.class.getClassLoader()).getURLs();
	}

	private static void addClassPathEntry(URL url) {
		URLClassLoader loader = (URLClassLoader) WhereIsThatStaticImport.class.getClassLoader();

		try {
			Method method = URLClassLoader.class.getDeclaredMethod("addURL", URL.class);
			method.setAccessible(true);
			method.invoke(loader, new Object[] { url });
		} catch (Throwable t) {
			t.printStackTrace();
			throw new IllegalStateException("Error, could not add URL to classloader");
		}
	}

	private final Map<String, List<Pair<Class<?>, Method>>> methodLookup;
	private final Map<String, List<Pair<Class<?>, Field>>> fieldLookup;

	public WhereIsThatStaticImport() {
		methodLookup = new HashMap<>();
		fieldLookup = new HashMap<>();
	}

	public List<Pair<Class<?>, Method>> queryMethod(String query, int maxResults) {
		Set<String> found = new HashSet<>();
		List<Pair<Class<?>, Method>> result = new ArrayList<>();

		String caps = identifierToCaps(query, IdentifierType.METHOD);

		List<Pair<Class<?>, Method>> matches = methodLookup.get(caps);
		if (matches == null) {
			return result;
		}

		List<String> queryParts = identifierToParts(query, IdentifierType.METHOD);

		outer: for (Pair<Class<?>, Method> match : matches) {
			String key = match.first().getName() + ":" + match.second().getName();
			if (!found.add(key)) {
				continue;
			}

			Method method = match.second();
			String methodName = method.getName();

			List<String> nameParts = identifierToParts(methodName, IdentifierType.METHOD);
			for (int i = 0; i < queryParts.size(); i++) {
				String queryPart = queryParts.get(i);
				String namePart = nameParts.get(i);
				if (!namePart.startsWith(queryPart)) {
					continue outer;
				}
			}

			result.add(match);
		}

		Collections.sort(result, new Comparator<Pair<Class<?>, Method>>() {
			@Override
			public int compare(Pair<Class<?>, Method> o1, Pair<Class<?>, Method> o2) {
				int comp = compareLengthThenContent(o1.first().getSimpleName(), o2.first().getSimpleName());
				if (comp == 0) {
					comp = compareLengthThenContent(o1.second().getName(), o2.second().getName());
				}
				return comp;
			}

			private int compareLengthThenContent(String s1, String s2) {
				int comp = s1.length() - s2.length();
				if (comp == 0) {
					comp = s1.compareTo(s2);
				}
				return comp;
			}
		});

		if (maxResults > 0 && result.size() > maxResults) {
			result = result.subList(0, maxResults);
		}

		return result;
	}

	public List<Pair<Class<?>, Field>> queryField(String query, int maxResults) {
		Set<String> found = new HashSet<>();
		List<Pair<Class<?>, Field>> result = new ArrayList<>();

		String caps = identifierToCaps(query, IdentifierType.FIELD);

		List<Pair<Class<?>, Field>> matches = fieldLookup.get(caps);
		if (matches == null) {
			return result;
		}

		List<String> queryParts = identifierToParts(query, IdentifierType.FIELD);

		outer: for (Pair<Class<?>, Field> match : matches) {
			String key = match.first().getName() + ":" + match.second().getName();
			if (!found.add(key)) {
				continue;
			}

			Field field = match.second();
			String fieldName = field.getName();

			List<String> nameParts = identifierToParts(fieldName, IdentifierType.FIELD);
			for (int i = 0; i < queryParts.size(); i++) {
				String queryPart = queryParts.get(i);
				String namePart = nameParts.get(i);
				if (!namePart.startsWith(queryPart)) {
					continue outer;
				}
			}

			result.add(match);
		}

		Collections.sort(result, new Comparator<Pair<Class<?>, Field>>() {
			@Override
			public int compare(Pair<Class<?>, Field> o1, Pair<Class<?>, Field> o2) {
				int comp = compareLengthThenContent(o1.first().getSimpleName(), o2.first().getSimpleName());
				if (comp == 0) {
					comp = compareLengthThenContent(o1.second().getName(), o2.second().getName());
				}
				return comp;
			}

			private int compareLengthThenContent(String s1, String s2) {
				int comp = s1.length() - s2.length();
				if (comp == 0) {
					comp = s1.compareTo(s2);
				}
				return comp;
			}
		});

		if (maxResults > 0 && result.size() > maxResults) {
			result = result.subList(0, maxResults);
		}

		return result;
	}

	private void discoverJar(InputStream in, String entryPrefix) throws IOException {
		JarInputStream jis = new JarInputStream(in);
		for (JarEntry entry; (entry = jis.getNextJarEntry()) != null;) {
			discoverJarEntry(entry, entryPrefix);
		}
	}

	private void discoverJarEntry(JarEntry entry, String entryPrefix) {
		String entryName = entry.getName();
		if (!entryName.endsWith(".class")) {
			return;
		}
		if (!entryName.startsWith(entryPrefix)) {
			return;
		}

		String className = entryName.substring(0, entryName.length() - ".class".length()).replace('/', '.');

		Class<?> clazz;
		try {
			clazz = Class.forName(className);
		} catch (Throwable exc) {
			System.err.println("warning: " + exc.getMessage());
			return;
		}

		this.discoverClass(clazz);
	}

	private void discoverClass(Class<?> clazz) {
		for (Method method : clazz.getMethods()) {
			if (!Modifier.isStatic(method.getModifiers())) {
				continue;
			}

			String name = method.getName();

			for (String caps = identifierToCaps(name, IdentifierType.METHOD); !caps.isEmpty(); caps = caps.substring(0, caps.length() - 1)) {
				List<Pair<Class<?>, Method>> matches = methodLookup.get(caps);
				if (matches == null) {
					methodLookup.put(caps, matches = new ArrayList<>());
				}

				matches.add(new Pair<Class<?>, Method>(clazz, method));
			}
		}

		for (Field field : clazz.getFields()) {
			if (!Modifier.isStatic(field.getModifiers())) {
				continue;
			}

			String name = field.getName();

			for (String caps = identifierToCaps(name, IdentifierType.FIELD); !caps.isEmpty(); caps = caps.substring(0, caps.length() - 1)) {
				List<Pair<Class<?>, Field>> matches = fieldLookup.get(caps);
				if (matches == null) {
					fieldLookup.put(caps, matches = new ArrayList<>());
				}

				matches.add(new Pair<Class<?>, Field>(clazz, field));
			}
		}
	}

	private static String identifierToCaps(String name, IdentifierType qt) {
		if (name.isEmpty()) {
			return "";
		}
		StringBuilder caps = new StringBuilder();
		caps.append(Character.toUpperCase(name.charAt(0)));
		for (int i = 1; i < name.length(); i++) {
			char c = name.charAt(i);
			if (isSeperator(c, qt)) {
				caps.append(c);
			}
		}
		return caps.toString();
	}

	private static List<String> identifierToParts(String name, IdentifierType idType) {
		List<String> parts = new ArrayList<>();

		StringBuilder builder = new StringBuilder();

		for (int i = 0; i < name.length(); i++) {
			char c = name.charAt(i);
			if (isSeperator(c, idType)) {
				parts.add(builder.toString());
				builder.setLength(0);

				if (idType == IdentifierType.METHOD) {
					c = Character.toLowerCase(c);
				}
			}
			builder.append(c);
		}
		parts.add(builder.toString());

		return parts;
	}

	private static boolean isSeperator(char c, IdentifierType idType) {
		switch (idType) {
			case METHOD:
				return Character.isUpperCase(c) || Character.isDigit(c);

			case FIELD:
				return c == '_';

			default:
				throw new IllegalStateException();
		}
	}

	private static String getIdentifierName(Pair<Class<?>, ?> item) {
		return (item.second() instanceof Method) ? ((Method) item.second()).getName() : ((Field) item.second()).getName();
	}

	static enum IdentifierType {
		METHOD, FIELD
	}
}
